"""
Test suite for _HetznerClient
"""
import unittest
import requests

import requests_mock

from certbot_dns_hetzner.fakes import FAKE_API_TOKEN, FAKE_RECORD_RESPONSE, FAKE_DOMAIN, \
    FAKE_ZONES_RESPONSE_WITH_DOMAIN, FAKE_ZONES_RESPONSE_WITHOUT_DOMAIN, FAKE_RECORD_ID
from certbot_dns_hetzner.hetzner_client import HETZNER_API_ENDPOINT, _ZoneNotFoundException, \
    _MalformedResponseException, _NotAuthorizedException


class HetznerClientTest(unittest.TestCase):
    record_name = 'foo'
    record_content = 'bar'
    record_ttl = 42

    def setUp(self):
        from certbot_dns_hetzner.dns_hetzner import _HetznerClient  # pylint: disable=import-outside-toplevel
        self.client = _HetznerClient(FAKE_API_TOKEN)

    def test_add_record(self):
        with requests_mock.Mocker() as mock:
            mock.get('{0}/zones'.format(HETZNER_API_ENDPOINT), status_code=200, json=FAKE_ZONES_RESPONSE_WITH_DOMAIN)
            mock.post('{0}/records'.format(HETZNER_API_ENDPOINT), status_code=200, json=FAKE_RECORD_RESPONSE)
            response = self.client.add_record(FAKE_DOMAIN, "TXT", "somename", "somevalue", 42)
            self.assertEqual(response, FAKE_RECORD_RESPONSE)

    def test_add_record_but_zone_is_not_in_account(self):
        with requests_mock.Mocker() as mock:
            mock.get('{0}/zones'.format(HETZNER_API_ENDPOINT), status_code=200, json=FAKE_ZONES_RESPONSE_WITHOUT_DOMAIN)
            mock.post('{0}/records'.format(HETZNER_API_ENDPOINT), status_code=200, json=FAKE_RECORD_RESPONSE)
            self.assertRaises(
                _ZoneNotFoundException,
                self.client.add_record, FAKE_DOMAIN, "TXT", "somename", "somevalue", 42
            )

    def test_add_record_but_record_creation_not_200(self):
        with requests_mock.Mocker() as mock:
            mock.get('{0}/zones'.format(HETZNER_API_ENDPOINT), status_code=200, json=FAKE_ZONES_RESPONSE_WITH_DOMAIN)
            mock.post('{0}/records'.format(HETZNER_API_ENDPOINT), status_code=406)
            self.assertRaises(
                _MalformedResponseException,
                self.client.add_record, FAKE_DOMAIN, "TXT", "somename", "somevalue", 42
            )

    def test_add_record_but_zone_listing_is_401(self):
        with requests_mock.Mocker() as mock:
            mock.get('{0}/zones'.format(HETZNER_API_ENDPOINT), status_code=401)
            mock.post('{0}/records'.format(HETZNER_API_ENDPOINT), status_code=200)
            self.assertRaises(
                _NotAuthorizedException,
                self.client.add_record, FAKE_DOMAIN, "TXT", "somename", "somevalue", 42
            )

    def test_add_record_but_zone_listing_times_out(self):
        with requests_mock.Mocker() as mock:
            mock.get('{0}/zones'.format(HETZNER_API_ENDPOINT), exc=requests.ConnectTimeout)
            mock.post('{0}/records'.format(HETZNER_API_ENDPOINT), status_code=200)
            self.assertRaises(
                requests.ConnectionError,
                self.client.add_record, FAKE_DOMAIN, "TXT", "somename", "somevalue", 42
            )

    def test_delete_record(self):
        with requests_mock.Mocker() as mock:
            mock.delete('{0}/records/{1}'.format(HETZNER_API_ENDPOINT, FAKE_RECORD_ID), status_code=200)
            self.client.delete_record(FAKE_RECORD_ID)

    def test_delete_but_authorization_fails(self):
        with requests_mock.Mocker() as mock:
            mock.delete('{0}/records/{1}'.format(HETZNER_API_ENDPOINT, FAKE_RECORD_ID), status_code=401)
            self.assertRaises(
                _NotAuthorizedException,
                self.client.delete_record, FAKE_RECORD_ID
            )

    def test_delete_record_but_deletion_is_404(self):
        with requests_mock.Mocker() as mock:
            mock.delete('{0}/records/{1}'.format(HETZNER_API_ENDPOINT, FAKE_RECORD_ID), status_code=404)
            self.assertRaises(
                _MalformedResponseException,
                self.client.delete_record, FAKE_RECORD_ID
            )



"""
A Hetzner helper class to wrap the API relevant for the functionality in this plugin
"""
import json
import requests

HETZNER_API_ENDPOINT = 'https://dns.hetzner.com/api/v1'


class _HetznerException(Exception):
    pass


class _ZoneNotFoundException(_HetznerException):
    def __init__(self, domain_name, *args):
        super(_ZoneNotFoundException, self).__init__('Zone {0} not found in Hetzner account'.format(domain_name), *args)
        self.domain_name = domain_name


class _MalformedResponseException(_HetznerException):
    def __init__(self, cause, *args):
        super(_MalformedResponseException, self).__init__(
            'Received an unexpected response from Hetzner API:\n{0}'.format(cause), *args)
        self.cause = cause


class _RecordNotFoundException(_HetznerException):
    def __init__(self, record_name, *args):
        super(_RecordNotFoundException, self).__init__('Record with name {0} not found'.format(record_name), *args)
        self.record_name = record_name


class _NotAuthorizedException(_HetznerException):
    def __init__(self, *args):
        super(_NotAuthorizedException, self).__init__('Malformed authorization or invalid API token', *args)


class _HetznerClient:
    """
    A little helper class for operations on the Hetzner DNS API
    """

    def __init__(self, token):
        """
        Initialize client by providing a Hetzner DNS API token
        :param token: Hetzner DNS API Token retrieved from: https://dns.hetzner.com/settings/api-token
        """
        self.token = token

    @property
    def _headers(self):
        return {
            "Content-Type": "application/json",
            "X-Consumer-Username": "",
            "Auth-API-Token": self.token,
        }

    def add_record(self, domain, record_type, name, value, ttl):  # pylint: disable=too-many-arguments
        """
        API call to add record to zone matching ``domain`` to your Hetzner Account, while specifying ``record_type``,
        ``name``, ``value`` and ``ttl``
        :param domain: Domain to determine zone where record should be added
        :param record_type: A valid DNS record type
        :param name: Full record name
        :param value: Record value
        :param ttl: Time to live
        :raises ._MalformedResponseException: If the response is missing expected values or is invalid JSON
        :raises ._ZoneNotFoundException: If no zone with the SLD and TLD of ``domain`` is found in your Hetzner account
        :raises ._NotAuthorizedException: If Hetzner does not accept the authorization credentials
        :raises requests.exceptions.ConnectionError: If the API request fails
        """
        zone_id = self._get_zone_id_by_domain(domain)
        create_record_response = requests.post(
            url="{0}/records".format(HETZNER_API_ENDPOINT),
            headers=self._headers,
            data=json.dumps({
                "value": value,
                "ttl": ttl,
                "type": record_type,
                "name": name,
                "zone_id": zone_id
            })
        )
        if create_record_response.status_code == 401:
            raise _NotAuthorizedException()
        try:
            return create_record_response.json()
        except (ValueError, UnicodeDecodeError) as exception:
            raise _MalformedResponseException(exception)

    def delete_record_by_name(self, domain, record_name):
        """
        Searches for a zone matching ``domain``, if found find and delete a record matching ``record_name`
        Deletes record with ``record_id`` from your Hetzner Account
        :param domain: Domain of zone in which the record should be found
        :param record_name: ID of record to be deleted
        :raises requests.exceptions.ConnectionError: If the API request fails
        :raises ._MalformedResponseException: If the API response is not 200
        :raises ._ZoneNotFoundException: If no zone is found matching ``domain``
        :raises ._RecordNotFoundException: If no record is found matching ``record_name``
        :raises ._NotAuthorizedException: If Hetzner does not accept the authorization credentials
        """
        zone_id = self._get_zone_id_by_domain(domain)
        record_id = self._get_record_id_by_name(zone_id, record_name)
        self.delete_record(record_id)

    def delete_record(self, record_id):
        """
        Deletes record with ``record_id`` from your Hetzner Account
        :param record_id: ID of record to be deleted
        :raises requests.exceptions.ConnectionError: If the API request fails
        :raises ._MalformedResponseException: If the API response is not 200
        :raises ._NotAuthorizedException: If Hetzner does not accept the authorization credentials
        """
        response = requests.delete(
            url="{0}/records/{1}".format(HETZNER_API_ENDPOINT, record_id),
            headers=self._headers
        )
        if response.status_code == 401:
            raise _NotAuthorizedException()
        if response.status_code != 200:
            raise _MalformedResponseException('Status code not 200')

    def _get_record_id_by_name(self, zone_id, record_name):
        """
        :param zone_id: ID of dns zone where the record should be searched
        :param record_name: Name of the record that is searched
        :return: The ID of the record with name ``record_name`` if found
        :raises ._MalformedResponseException: If the response is missing expected values or is invalid JSON
        :raises requests.exceptions.ConnectionError: If the API request fails
        :raises ._NotAuthorizedException: If Hetzner does not accept the authorization credentials
        :rtype: str
        """
        records_response = requests.get(
            url="{0}/records".format(HETZNER_API_ENDPOINT),
            params={
                'zone_id': zone_id,
            },
            headers=self._headers
        )
        if records_response.status_code == 401:
            raise _NotAuthorizedException()
        try:
            records = records_response.json()['records']
            for record in records:
                if record['name'] == record_name:
                    return record['id']
        except (ValueError, UnicodeDecodeError, KeyError) as exception:
            raise _MalformedResponseException(exception)
        raise _RecordNotFoundException(record_name)

    def _get_zone_id_by_domain(self, domain):
        """
        Requests all dns zones from your Hetzner account and searches for a specific one to determine the ID of it
        :param domain: Name of dns zone where the record should be searched
        :return: The ID of the zone that is SLD and TLD of ``domain`` - if found
        :raises ._MalformedResponseException: If the response is missing expected values or is invalid JSON
        :raises ._ZoneNotFoundException: If no zone with the SLD and TLD of ``domain`` is found in your Hetzner account
        :raises ._NotAuthorizedException: If Hetzner does not accept the authorization credentials
        :raises requests.exceptions.ConnectionError: If the API request fails
        :rtype: str
        """
        domain_tokens = domain.split('.')
        zones_response = requests.get(
            url="{0}/zones".format(HETZNER_API_ENDPOINT),
            headers=self._headers,
        )
        if zones_response.status_code == 401:
            raise _NotAuthorizedException()
        try:
            zones = zones_response.json()['zones']
            for zone in zones:
                zone_name_tokens = zone['name'].split('.')
                # take sld and tld to match zones
                if zone_name_tokens[-1] == domain_tokens[-1] and zone_name_tokens[-2] == domain_tokens[-2]:
                    return zone['id']
        except (KeyError, UnicodeDecodeError, ValueError) as exception:
            raise _MalformedResponseException(exception)
        raise _ZoneNotFoundException(domain)



"""
Fakes needed for tests
"""

FAKE_API_TOKEN = 'XXXXXXXXXXXXXXXXXXXxxx'
FAKE_RECORD = {
    "record": {
        'id': "123Fake",
    }
}

FAKE_DOMAIN = 'some.domain'
FAKE_ZONE_ID = 'xyz'
FAKE_RECORD_ID = 'zzz'

FAKE_RECORD_RESPONSE = {
    "record": {
        "id": FAKE_RECORD_ID,
        "name": "string",
    }
}

FAKE_ZONES_RESPONSE_WITH_DOMAIN = {
    "zones": [
        {
            "id": FAKE_ZONE_ID,
            "name": FAKE_DOMAIN,
        }
    ]
}

FAKE_ZONES_RESPONSE_WITHOUT_DOMAIN = {
    "zones": [
        {
            "id": "string",
            "name": "string",
        }
    ]
}



"""DNS Authenticator for Hetzner DNS."""
import requests

import zope.interface

from certbot import errors
from certbot import interfaces
from certbot.plugins import dns_common

from certbot_dns_hetzner.hetzner_client import \
    _MalformedResponseException, \
    _HetznerClient, \
    _ZoneNotFoundException, _NotAuthorizedException

TTL = 60


@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for Hetzner
    This Authenticator uses the Hetzner DNS API to fulfill a dns-01 challenge.
    """

    description = 'Obtain certificates using a DNS TXT record (if you are using Hetzner for DNS).'

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self.credentials = None

    @classmethod
    def add_parser_arguments(cls, add):  # pylint: disable=arguments-differ
        super(Authenticator, cls).add_parser_arguments(add, default_propagation_seconds=60)
        add('credentials', help='Hetzner credentials INI file.')

    def more_info(self):  # pylint: disable=missing-function-docstring,no-self-use
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
               'the Hetzner API.'

    def _setup_credentials(self):
        self.credentials = self._configure_credentials(
            'credentials',
            'Hetzner credentials INI file',
            {
                'api_token': 'Hetzner API Token from \'https://dns.hetzner.com/settings/api-token\'',
            }
        )

    def _perform(self, domain, validation_name, validation):
        try:
            self._get_hetzner_client().add_record(
                domain,
                "TXT",
                self._fqdn_format(validation_name),
                validation,
                TTL
            )
        except (
                _ZoneNotFoundException,
                requests.ConnectionError,
                _MalformedResponseException,
                _NotAuthorizedException
        ) as exception:
            raise errors.PluginError(exception)

    def _cleanup(self, domain, validation_name, validation):
        try:
            self._get_hetzner_client().delete_record_by_name(domain, self._fqdn_format(validation_name))
        except (requests.ConnectionError, _NotAuthorizedException) as exception:
            raise errors.PluginError(exception)

    def _get_hetzner_client(self):
        return _HetznerClient(
            self.credentials.conf('api_token'),
        )

    def _fqdn_format(self, name):
        if not name.endswith('.'):
            return '{0}.'.format(name)
        else:
            return name



"""Tests for certbot_dns_ispconfig.dns_ispconfig."""

import unittest

import mock

from certbot.compat import os
from certbot.errors import PluginError
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import DOMAIN
from certbot.tests import util as test_util

from certbot_dns_hetzner.fakes import FAKE_API_TOKEN, FAKE_RECORD
from certbot_dns_hetzner.hetzner_client import _ZoneNotFoundException


class AuthenticatorTest(
        test_util.TempDirTestCase,
        dns_test_common.BaseAuthenticatorTest
):
    """
    Test for Hetzner DNS Authenticator
    """
    def setUp(self):
        super(AuthenticatorTest, self).setUp()
        from certbot_dns_hetzner.dns_hetzner import Authenticator  # pylint: disable=import-outside-toplevel

        path = os.path.join(self.tempdir, 'fake_credentials.ini')
        dns_test_common.write(
            {
                'hetzner_api_token': FAKE_API_TOKEN,
            },
            path,
        )

        super(AuthenticatorTest, self).setUp()
        self.config = mock.MagicMock(
            hetzner_credentials=path, hetzner_propagation_seconds=0
        )  # don't wait during tests

        self.auth = Authenticator(self.config, 'hetzner')

        self.mock_client = mock.MagicMock()
        # _get_ispconfig_client | pylint: disable=protected-access
        self.auth._get_hetzner_client = mock.MagicMock(return_value=self.mock_client)

    def test_perform(self):
        self.mock_client.add_record.return_value = FAKE_RECORD
        self.auth.perform([self.achall])
        self.mock_client.add_record.assert_called_with(
            DOMAIN, 'TXT', '_acme-challenge.' + DOMAIN + '.', mock.ANY, mock.ANY
        )
        self.assertEqual(self.auth.record_id, FAKE_RECORD['record']['id'])

    def test_perform_but_raises_zone_not_found(self):
        self.mock_client.add_record.side_effect = mock.MagicMock(side_effect=_ZoneNotFoundException(DOMAIN))
        self.assertRaises(
            PluginError,
            self.auth.perform, [self.achall]
        )
        self.mock_client.add_record.assert_called_with(
            DOMAIN, 'TXT', '_acme-challenge.' + DOMAIN + '.', mock.ANY, mock.ANY
        )

    def test_cleanup(self):
        self.mock_client.add_record.return_value = FAKE_RECORD
        # _attempt_cleanup | pylint: disable=protected-access
        self.auth.perform([self.achall])
        self.auth._attempt_cleanup = True
        self.auth.cleanup([self.achall])

        self.mock_client.delete_record.assert_called_with(record_id=FAKE_RECORD['record']['id'])


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



import os

from setuptools import find_packages
from setuptools import setup

version = '1.0.2'

# This package relies on PyOpenSSL, requests, and six, however, it isn't
# specified here to avoid masking the more specific request requirements in
# acme. See https://github.com/pypa/pip/issues/988 for more info.
install_requires = [
    'certbot>=1.1.0',
    'setuptools',
    'zope.interface',
    'requests',
    'mock',
    'requests-mock',
]

BASE_PATH = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(BASE_PATH, "README.md")) as f:
    long_description = f.read()

setup(
    name='certbot-dns-hetzner',
    version=version,
    description="Hetzner DNS Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ctrlaltcoop/certbot-dns-hetzner',
    author="ctrl.alt.coop",
    author_email='kontakt@ctrl.alt.coop',
    license='Apache License 2.0',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        'certbot.plugins': [
            'dns-hetzner = certbot_dns_hetzner.dns_hetzner:Authenticator',
        ],
    },
    test_suite="certbot_dns_ispconfig",
)
